/*** * **** * ***** ********* ** ****** ***** *** ***** ********
**
** Copyright (c) 2005-2009, SkeTo Project
** All rights reserved.
**
** This file is part of the SkeTo Library. This library is free
** software; you can redistribute it and/or modify it under the
** terms of the MIT License (see COPYING in the package).
** 
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
**
*** * **** * ***** ********* ** ****** ***** *** ***** ********/

/**
 * @file list_skeletons_noopt.tpp
 * @brief Template implementation of list skeletons without fusion optimization.
 */
#ifndef __LIST_SKELETONS_NOOPT_TPP__
#define __LIST_SKELETONS_NOOPT_TPP__
#include <cassert>

#include "list_skeletons_common.tpp"

namespace sketo {
  //------------------------------------------------------------------------
  // generate
  template <typename F>
  dist_list<typename F::result_type>
  list_skeletons::generate(int global_size, const F& f)
  {
    return _impl::ls_adpt::generate<F, typename F::result_type>(global_size, f);
  }

  template <typename F, typename A>
  dist_list<A>
  _impl::ls_adpt::generate(int global_size,
			   const F& f)
  {
    dist_list<A> ret(global_size);
    
    A* ret_it = ret.data->buf;
    A* const ret_it_end = ret.data->buf + ret.data->local_size;
    int index = ret.data->local_begin;
    for (;ret_it != ret_it_end; ++index, ++ret_it) {
      *ret_it = f(index);
    }
    
    return ret;
  }

  //------------------------------------------------------------------------
  // map
  template <typename F, typename A>
  dist_list<typename F::result_type>
  list_skeletons::map(F f,
		      const dist_list<A>& as)
  {
    return _impl::ls_adpt::map<F, typename F::argument_type, typename F::result_type>(f, as);
  }

  template <typename F, typename A, typename B>
  dist_list<B>
  _impl::ls_adpt::map(const F& f,
		      const dist_list<A>& as)
  {
    dist_list<B> ret(as.get_global_size());

    B* const ret_buf = ret.data->buf;
    const int n = as.get_local_size();
    for (int i = 0; i < n; ++i) {
      ret_buf[i] = f(as.local_get(i));
    }
    
    return ret;
  }

  //------------------------------------------------------------------------
  // map_with_index
  template <typename F, typename A>
  dist_list<typename F::result_type>
  list_skeletons::map_with_index(const F& f,
				 const dist_list<A>& as)
  {
    return _impl::ls_adpt::map_with_index<F, A, typename F::result_type>(f, as);
  }

  template <typename F, typename A, typename B>
  dist_list<B>
  _impl::ls_adpt::map_with_index(const F& f,
				 const dist_list<A>& as)
  {
    dist_list<B> ret(as.get_global_size());

    B* const ret_buf = ret.data->buf;
    const int n = as.get_local_size();
    for (int i = 0, index = ret.data->local_begin; i < n; ++i, ++index) {
      ret_buf[i] = f(index, as.local_get(i));
    }
    
    return ret;
  }
    
  //------------------------------------------------------------------------
  // zip and zipwith
  template <typename A, typename B>
  dist_list<std::pair<A, B> >
  list_skeletons::zip(const dist_list<A>& as,
		      const dist_list<B>& bs)
  {
    return _impl::ls_adpt::zipwith<sketo::functions::mkpair<A, B>, A, B, std::pair<A, B> >(sketo::functions::mkpair<A, B>(), as, bs);
  }

  template <typename F, typename A, typename B>
  dist_list<typename F::result_type>
  list_skeletons::zipwith(const F& f,
			  const dist_list<A>& as,
			  const dist_list<B>& bs)
  {
    return _impl::ls_adpt::zipwith<F, A, B, typename F::result_type>(f, as, bs);
  }

  template <typename F, typename A, typename B, typename C>
  dist_list<C>
  _impl::ls_adpt::zipwith(const F& f,
			  const dist_list<A>& as,
			  const dist_list<B>& bs)
  {
    assert(as.get_global_size() == bs.get_global_size());
    dist_list<C> ret(as.get_global_size());

    C* const ret_buf = ret.data->buf;
    const int n = as.get_local_size();
    for (int i = 0; i < n; ++i) {
      ret_buf[i] = f(as.local_get(i), bs.local_get(i));
    }
    
    return ret;
  }

  //------------------------------------------------------------------------
  // reduce
  template <typename OPLUS>
  typename OPLUS::result_type
  list_skeletons::reduce(const OPLUS &oplus,
			 const dist_list<typename OPLUS::first_argument_type>& as)
  {
    return _impl::ls_adpt::reduce(oplus, as);
  }

  template <typename OPLUS, typename A>
  A
  _impl::ls_adpt::reduce(const OPLUS &oplus,
			 const dist_list<A> &as)
  {
    // local reduce
    A val = as.local_get(0);
    const int n = as.get_local_size();
    for (int i = 1; i < n; ++i) {
      val = oplus(val, as.local_get(i));
    }
    
    // global reduce
    _impl::ls_adpt::global_reduce(oplus, &val);
    return val;
  }

  //------------------------------------------------------------------------
  // scan
  template <typename OPLUS>
  dist_list<typename OPLUS::result_type>
  list_skeletons::scan(const OPLUS &oplus,
		       typename OPLUS::first_argument_type e,
		       const dist_list<typename OPLUS::second_argument_type>& as,
		       typename OPLUS::result_type *last_val)
  {
    return _impl::ls_adpt::scan<OPLUS, typename OPLUS::result_type>(oplus, e, as, last_val);
  }

  template <typename OPLUS, typename A>
  dist_list<A>
  _impl::ls_adpt::scan(const OPLUS& oplus,
		       A e,
		       const dist_list<A> &as,
		       A* last_val)
  {
    // local reduce
    A total_sum;
    {
      const int n = as.get_local_size();
      total_sum = as.local_get(0);
      for (int i = 1; i < n; ++i) {
	total_sum = oplus(total_sum, as.local_get(i));
      }
    }

    // global scan
    A first_val;
    global_scan(oplus, &first_val, &total_sum);
    if (rank == 0) {
      first_val = e;
    } else {
      first_val = oplus(e, first_val);
    }
    
    // share the last value (= total_sum)
    if (last_val) {
      if (sketo::rank == 0) *last_val = oplus(e, total_sum);
      MPI_Bcast(last_val, sizeof(A), MPI_BYTE, 0, MPI_COMM_WORLD);
    }

    dist_list<A> bs(as.get_global_size());
    // local scan
    {
      A* const bs_buf = bs.data->buf;
      const int n = as.get_local_size();
      bs_buf[0] = first_val;
      for (int i = 1; i < n; ++i) {
	bs_buf[i] = oplus(bs_buf[i-1], as.local_get(i-1));
      }
    }
    
    return bs;
  }

  //------------------------------------------------------------------------
  // scanr
  template <typename OPLUS>
  dist_list<typename OPLUS::result_type>
  list_skeletons::scanr(const OPLUS &oplus,
			typename OPLUS::second_argument_type e,
			const dist_list<typename OPLUS::first_argument_type>& as,
			typename OPLUS::result_type *first_val)
  {
    return _impl::ls_adpt::scanr<OPLUS, typename OPLUS::result_type>(oplus, e, as, first_val);
  }

  template <typename OPLUS, typename A>
  dist_list<A>
  _impl::ls_adpt::scanr(const OPLUS& oplus,
			A e,
			const dist_list<A> &as,
			A* first_val)
  {
    // local reduce
    A total_sum;
    {
      const int n = as.get_local_size();
      total_sum = as.local_get(0);
      for (int i = 1; i < n; ++i) {
	total_sum = oplus(total_sum, as.local_get(i));
      }
    }
    
    // global scanr
    A last_val;
    global_scanr(oplus, &last_val, &total_sum);
    if (rank == procs - 1) {
      last_val = e;
    } else {
      last_val = oplus(last_val, e);
    }
    
    // share the first value (= total_sum)
    if (first_val) {
      if (sketo::rank == procs - 1) *first_val = oplus(total_sum, e);
      MPI_Bcast(first_val, sizeof(A), MPI_BYTE, procs-1, MPI_COMM_WORLD);
    }

    dist_list<A> bs(as.get_global_size());
    // local scan
    {
      A* const bs_buf = bs.data->buf;
      bs_buf[as.get_local_size() - 1] = last_val;
      for (int i = as.get_local_size() - 2; i >= 0; --i) {
	bs_buf[i] = oplus(as.local_get(i+1), bs_buf[i+1]);
      }
    }
    
    return bs;
  }
  
  //------------------------------------------------------------------------
  // postscan
  template <typename OPLUS>
  dist_list<typename OPLUS::result_type>
  list_skeletons::postscan(const OPLUS &oplus,
			   const dist_list<typename OPLUS::first_argument_type>& as)
  {
    return _impl::ls_adpt::postscan<OPLUS, typename OPLUS::result_type>(oplus, as);
  }

  //------------------------------------------------------------------------
  // postscanr
  template <typename OPLUS>
  dist_list<typename OPLUS::result_type>
  list_skeletons::postscanr(const OPLUS &oplus,
			    const dist_list<typename OPLUS::second_argument_type>& as)
  {
    return _impl::ls_adpt::postscanr<OPLUS, typename OPLUS::result_type>(oplus, as);
  }

  

  //------------------------------------------------------------------------
  // shiftl (shift to left)
  template <typename A>
  dist_list<A>
  list_skeletons::shiftl(const A &rightmost,
			 const dist_list<A> &as,
			 A *leftmost)
  {
    return _impl::ls_adpt::shiftl(rightmost, as, leftmost);
  }

  template <typename A>
  dist_list<A>
  _impl::ls_adpt::shiftl(const A &rightmost,
			 const dist_list<A> &as,
			 A *leftmost)
  {
    dist_list<A> ret(as.get_global_size());
    A real_leftmost = as.local_get(0);

    MPI_Request reqs, reqr;
    // sending or receiving edges
    if (sketo::rank != 0) {
      MPI_Isend(&real_leftmost, sizeof(A), MPI_BYTE, sketo::rank - 1,
		TAG_SHIFTL, MPI_COMM_WORLD, &reqs);
    }
    
    if (sketo::rank == sketo::procs - 1) {
      ret.data->buf[ret.data->local_size - 1] = rightmost;
    } else {
      MPI_Irecv(ret.data->buf + ret.data->local_size - 1, sizeof(A), MPI_BYTE, sketo::rank + 1,
		TAG_SHIFTL, MPI_COMM_WORLD, &reqr);
    }

    // moving elements
    {
      A* const ret_buf = ret.data->buf;
      const int n = as.get_local_size();
      for (int i = 0; i < n - 1; ++i) {
	ret_buf[i] = as.local_get(i+1);
      }
    }

    // wait sening or receiving
    if (sketo::rank != 0) {
      MPI_Wait(&reqs, MPI_STATUS_IGNORE);
    }

    if (sketo::rank == sketo::procs - 1) {
      // do nothing
    } else {
      MPI_Wait(&reqr, MPI_STATUS_IGNORE);
    }

    // share leftmost if specified
    if (leftmost) {
      *leftmost = real_leftmost;
      MPI_Bcast(leftmost, sizeof(A), MPI_BYTE, 0, MPI_COMM_WORLD);
    }
    return ret;
  }


  //------------------------------------------------------------------------
  // shiftr (shift to right)
  template <typename A>
  dist_list<A>
  list_skeletons::shiftr(const A &leftmost,
			 const dist_list<A> &as,
			 A* rightmost)
  {
    return _impl::ls_adpt::shiftr(leftmost, as, rightmost);
  }

  template <typename A>
  dist_list<A>
  _impl::ls_adpt::shiftr(const A &leftmost,
			 const dist_list<A> &as,
			 A *rightmost)
  {
    dist_list<A> ret(as.get_global_size());
    A real_rightmost = as.local_get(as.get_local_size()-1);

    MPI_Request reqs, reqr;
    // sending or receiving edge
    if (sketo::rank != sketo::procs - 1) {
      MPI_Isend(&real_rightmost, sizeof(A), MPI_BYTE, sketo::rank + 1,
		TAG_SHIFTR, MPI_COMM_WORLD, &reqs);
    }

    if (sketo::rank == 0) {
      ret.data->buf[0] = leftmost;
    } else {
      MPI_Irecv(ret.data->buf, sizeof(A), MPI_BYTE, sketo::rank - 1,
		TAG_SHIFTR, MPI_COMM_WORLD, &reqr);
    }
      
    // moving elements
    {
      A* const it_ret = ret.data->buf;
      const int n = as.get_local_size();
      for (int i = 1; i < n; ++i) {
	it_ret[i] = as.local_get(i-1);
      }
    }

    // wait sening or receiving
    if (sketo::rank != sketo::procs - 1) {
      MPI_Wait(&reqs, MPI_STATUS_IGNORE);
    }
    
    if (sketo::rank == 0) {
      // do nothing
    } else {
      MPI_Wait(&reqr, MPI_STATUS_IGNORE);
    }

    // share leftmost if specified
    if (rightmost) {
      *rightmost = real_rightmost;
      MPI_Bcast(rightmost, sizeof(A), MPI_BYTE, procs - 1, MPI_COMM_WORLD);
    }

    return ret;
  }

  //------------------------------------------------------------------------
  // gscanl
  template <typename F, typename G, typename OMINUS, typename OPLUS>
  dist_list<typename F::result_type>
  list_skeletons::gscanl(const F &f,
			 const typename F::first_argument_type &c,
			 const G &g,
			 const OMINUS &ominus,
			 const OPLUS &oplus,
			 const dist_list<typename F::second_argument_type>& as)
  {
    return _impl::ls_adpt::gscanl<F, G, OMINUS, OPLUS, typename F::second_argument_type, typename F::first_argument_type, typename G::result_type>(f, c, g, ominus, oplus, as);
  }


  //------------------------------------------------------------------------
  // gscanr
  template <typename F, typename G, typename OMINUS, typename OPLUS>
  dist_list<typename F::result_type>
  list_skeletons::gscanr(const F &f,
			 const typename F::first_argument_type &c,
			 const G &g,
			 const OMINUS &ominus,
			 const OPLUS &oplus,
			 const dist_list<typename F::second_argument_type>& as)
  {
    return _impl::ls_adpt::gscanr<F, G, OMINUS, OPLUS, typename F::second_argument_type, typename F::first_argument_type, typename G::result_type>(f, c, g, ominus, oplus, as);
  }

}
  
#endif // __LIST_SKELETONS_NOOPT_TPP__
